package com.sunyard.chsm.service.impl;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.sunyard.chsm.constant.ParamConfKeyConstant;
import com.sunyard.chsm.dto.DeviceDTO;
import com.sunyard.chsm.dto.TmkStatus;
import com.sunyard.chsm.enums.ManufacturerEnum;
import com.sunyard.chsm.enums.ManufacturerModelEnum;
import com.sunyard.chsm.enums.TmkStatusEnum;
import com.sunyard.chsm.mapper.ParamConfMapper;
import com.sunyard.chsm.mapper.SpDeviceMapper;
import com.sunyard.chsm.mapper.TmkInfoMapper;
import com.sunyard.chsm.model.entity.Device;
import com.sunyard.chsm.model.entity.ParamConf;
import com.sunyard.chsm.sdf.adapter.SdfApiAdapter;
import com.sunyard.chsm.sdf.adapter.SdfApiAdapterFactory;
import com.sunyard.chsm.sdf.context.AlgId;
import com.sunyard.chsm.sdf.model.EccCipher;
import com.sunyard.chsm.sdf.model.EccPubKey;
import com.sunyard.chsm.service.DeviceService;
import com.sunyard.chsm.utils.CodecUtils;
import lombok.extern.slf4j.Slf4j;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.springframework.beans.BeanUtils;
import org.springframework.data.util.Pair;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;
import org.springframework.util.CollectionUtils;
import org.springframework.util.StringUtils;

import javax.annotation.Resource;
import java.time.LocalDateTime;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * @author liulu
 * @since 2024/10/17
 */
@Slf4j
@Service("sp_DeviceServiceImpl")
@Transactional
public class DeviceServiceImpl implements DeviceService {

    @Resource
    private SpDeviceMapper spDeviceMapper;
    @Resource
    private ParamConfMapper paramConfMapper;
    @Resource
    private TmkInfoMapper tmkInfoMapper;

    @Override
    public Page<DeviceDTO.DeviceView> selectPageList(DeviceDTO.Query query) {
        LambdaQueryWrapper<Device> wrapper = new LambdaQueryWrapper<Device>()
                .eq(StringUtils.hasText(query.getManufacturers()), Device::getManufacturer, query.getManufacturers())
                .eq(StringUtils.hasText(query.getModel()), Device::getManufacturerModel, query.getModel())
                .like(StringUtils.hasText(query.getName()), Device::getName, query.getName())
                .orderByDesc(Device::getCreateTime);

        IPage<Device> page = spDeviceMapper.selectPage(
                new Page<>(query.getPageNumber(), query.getPageSize()),
                wrapper);
        List<Device> records = page.getRecords();
        if (CollectionUtils.isEmpty(records)) {
            return new Page<>(page.getCurrent(), page.getSize(), page.getTotal());
        }
        List<DeviceDTO.DeviceView> viewList = records.stream()
                .map(it -> {
                    DeviceDTO.DeviceView view = new DeviceDTO.DeviceView();
                    BeanUtils.copyProperties(it, view);
                    view.setIsGrouped(it.getGroupId() > 0);
                    Optional.ofNullable(ManufacturerEnum.of(it.getManufacturer()))
                            .map(ManufacturerEnum::getName)
                            .ifPresent(view::setManufacturerText);
                    Optional.ofNullable(ManufacturerModelEnum.of(it.getManufacturerModel()))
                            .map(ManufacturerModelEnum::getName)
                            .ifPresent(view::setManufacturerModelText);
                    return view;
                })
                .collect(Collectors.toList());

        return new Page<DeviceDTO.DeviceView>(page.getCurrent(), page.getSize(), page.getTotal()).setRecords(viewList);
    }

    @Override
    public List<DeviceDTO.DeviceView> getUngroupedList() {

        LambdaQueryWrapper<Device> wrapper = new LambdaQueryWrapper<Device>()
                .eq(Device::getGroupId, 0L);
        List<Device> devices = spDeviceMapper.selectList(wrapper);
        return devices.stream()
                .map(it -> {
                    DeviceDTO.DeviceView view = new DeviceDTO.DeviceView();
                    BeanUtils.copyProperties(it, view);
                    Optional.ofNullable(ManufacturerEnum.of(it.getManufacturer()))
                            .map(ManufacturerEnum::getName)
                            .ifPresent(view::setManufacturerText);
                    Optional.ofNullable(ManufacturerModelEnum.of(it.getManufacturerModel()))
                            .map(ManufacturerModelEnum::getName)
                            .ifPresent(view::setManufacturerModelText);
                    return view;
                })
                .collect(Collectors.toList());
    }

    @Override
    public DeviceDTO.Detail getDetailById(Long id) {
        Device exist = spDeviceMapper.selectById(id);
        Assert.notNull(exist, "查询的数据不存在");

        DeviceDTO.Detail detail = new DeviceDTO.Detail();
        BeanUtils.copyProperties(exist, detail);
        Optional.ofNullable(ManufacturerEnum.of(exist.getManufacturer()))
                .map(ManufacturerEnum::getName)
                .ifPresent(detail::setManufacturerText);
        Optional.ofNullable(ManufacturerModelEnum.of(exist.getManufacturerModel()))
                .map(ManufacturerModelEnum::getName)
                .ifPresent(detail::setManufacturerModelText);

        // todo
        detail.setCpuRate(0.5D);
        detail.setMemoryRate(0.5D);
        detail.setDiskRate(0.5D);

        return detail;
    }


    @Override
    public Long save(DeviceDTO.DeviceSave save) {
        ManufacturerEnum manufacturer = ManufacturerEnum.of(save.getManufacturer());
        Assert.notNull(manufacturer, "不支持的设备厂商: " + save.getManufacturer());
        ManufacturerModelEnum manufacturerModel = ManufacturerModelEnum.of(save.getManufacturerModel());
        Assert.notNull(manufacturerModel, "不支持的设备型号: " + save.getManufacturerModel());
        Assert.isTrue(manufacturerModel.getManufacturer() == manufacturer, "设备厂商和型号不匹配");

        checkName(save.getName());
        checkIpPort(save.getServiceIp(), save.getServicePort());

        Device device = new Device();
        BeanUtils.copyProperties(save, device);
        device.setId(IdWorker.getId());
        device.setGroupId(0L);
        device.setGroupName("");
        device.setWeight(1);
        device.setCreateTime(LocalDateTime.now());

        Pair<String, String> pair = checkDeviceSdf(save);
        device.setTmkStatus(StringUtils.hasText(pair.getSecond()) ? TmkStatusEnum.finished.name() : TmkStatusEnum.available.name());
        device.setPubKey(pair.getFirst());
        device.setEncTmk(pair.getSecond());

        spDeviceMapper.insert(device);
        return device.getId();
    }

    private void checkIpPort(String serviceIp, Integer servicePort) {
        LambdaQueryWrapper<Device> wrapper = new LambdaQueryWrapper<Device>()
                .eq(Device::getServiceIp, serviceIp)
                .eq(Device::getServicePort, servicePort);
        Device exist = spDeviceMapper.selectOne(wrapper);
        Assert.isNull(exist, "设备IP和端口已存在");
    }

    private Pair<String, String> checkDeviceSdf(DeviceDTO.DeviceSave save) {
        SdfApiAdapter sdfApiAdapter = SdfApiAdapterFactory.newInstance(save.getManufacturerModel(), save.getServiceIp(), save.getServicePort());

        String hd = sdfApiAdapter.openDevice();
        String hs = null;
        try {
            String encTmk = "";
            hs = sdfApiAdapter.openSession(hd);
            EccPubKey eccPubKey = sdfApiAdapter.exportEncPublicKeyECC(hs, save.getEncKeyIdx());
            if (Arrays.equals(eccPubKey.getPubKeyBytes(), new byte[64])) {
                throw new IllegalArgumentException("加密密钥索引不正确");
            }
            sdfApiAdapter.getPrivateKeyAccessRight(hs, save.getEncKeyIdx(), save.getAccessCredentials().getBytes());
            if (isTmkInit()) {
                Device device = getOneByStatus(TmkStatusEnum.finished);
                Assert.notNull(device, "系统主密钥设备异常");

                SdfApiAdapter tmkAdapter = SdfApiAdapterFactory.newInstance(save.getManufacturerModel(), save.getServiceIp(), save.getServicePort());
                String tmkHd = tmkAdapter.openDevice();
                String tmkHs = tmkAdapter.openSession(tmkHd);
                tmkAdapter.getPrivateKeyAccessRight(tmkHs, device.getEncKeyIdx(), device.getAccessCredentials().getBytes());
                EccCipher cipher = tmkAdapter.exchangeDigitEnvelopeBaseOnECC(tmkHs, device.getEncKeyIdx(), eccPubKey, EccCipher.fromHex(device.getEncTmk()));
                String hk = sdfApiAdapter.importKeyWithISKECC(hs, save.getEncKeyIdx(), cipher);
                sdfApiAdapter.destroyKey(hs, hk);
                encTmk = cipher.getC1C3C2Hex();
            } else {
                byte[] random = sdfApiAdapter.generateRandom(hs, 16);
                EccCipher eccCipher = sdfApiAdapter.externalEncryptECC(hs, eccPubKey, random);
                String hk = sdfApiAdapter.importKeyWithISKECC(hs, save.getEncKeyIdx(), eccCipher);
                sdfApiAdapter.destroyKey(hs, hk);
            }
            return Pair.of(eccPubKey.getPubKeyHex(), encTmk);
        } finally {
            Optional.ofNullable(hs).ifPresent(sdfApiAdapter::closeSession);
            sdfApiAdapter.closeDevice(hd);
        }
    }

    @Override
    public void update(DeviceDTO.DeviceSave update) {
        Assert.notNull(update.getId(), "id不能为空");
        Device exist = spDeviceMapper.selectById(update.getId());
        Assert.notNull(exist, "id对应的设备不存在");

        if (!Objects.equals(exist.getName(), update.getName())) {
            checkName(update.getName());
        }
        boolean ipChanged = !Objects.equals(exist.getServiceIp(), update.getServiceIp())
                || !Objects.equals(exist.getServicePort(), update.getServicePort());
        if (ipChanged) {
            checkIpPort(update.getServiceIp(), update.getServicePort());
        }

        Device up = new Device();
        BeanUtils.copyProperties(update, up);
        up.setUpdateTime(LocalDateTime.now());

        if (ipChanged
                || !Objects.equals(exist.getEncKeyIdx(), update.getEncKeyIdx())
                || !Objects.equals(exist.getAccessCredentials(), update.getAccessCredentials())) {
            Pair<String, String> pair = checkDeviceSdf(update);
            up.setTmkStatus(StringUtils.hasText(pair.getSecond()) ? TmkStatusEnum.finished.name() : TmkStatusEnum.available.name());
            up.setPubKey(pair.getFirst());
            up.setEncTmk(pair.getSecond());
        }
        spDeviceMapper.updateById(up);
    }

    @Override
    public void delete(Long id) {
        spDeviceMapper.deleteById(id);
    }

    @Override
    public TmkStatus getTMKStatus() {
        TmkStatus status = new TmkStatus();
        boolean init = isTmkInit();
        if (init) {
            status.setHasDevice(true);
            status.setTmkInit(true);
            return status;
        }
        status.setTmkInit(false);
        Device device = getOneByStatus(TmkStatusEnum.available);
        status.setHasDevice(device != null);
        return status;
    }

    @Override
    public void initTmk() {
        boolean tmkInit = isTmkInit();
        Assert.isTrue(!tmkInit, "主密钥已经初始化");

        Device device = getOneByStatus(TmkStatusEnum.available);
        Assert.notNull(device, "没有可以用于初始化主密钥的设备");
        SdfApiAdapter sdfApi = SdfApiAdapterFactory.newInstance(device.getManufacturerModel(), device.getServiceIp(), device.getServicePort());

        String hd = sdfApi.openDevice();
        String hs = sdfApi.openSession(hd);
        byte[] rk = sdfApi.generateRandom(hs, 16);
        EccPubKey pubKey = sdfApi.exportEncPublicKeyECC(hs, device.getEncKeyIdx());
        EccCipher cipher = sdfApi.externalEncryptECC(hs, pubKey, rk);
        sdfApi.getPrivateKeyAccessRight(hs, device.getEncKeyIdx(), device.getAccessCredentials().getBytes());
        String hk = sdfApi.importKeyWithISKECC(hs, device.getEncKeyIdx(), cipher);
        byte[] encrk = sdfApi.symEncrypt(hs, hk, AlgId.SGD_SM4_ECB, new byte[0], rk);
        byte[] prk = sdfApi.symDecrypt(hs, hk, AlgId.SGD_SM4_ECB, new byte[0], encrk);
        Assert.isTrue(Arrays.equals(rk, prk), "密码机加解密异常");

        sdfApi.destroyKey(hs, hk);
        sdfApi.closeSession(hs);
        sdfApi.closeDevice(hd);

        if (enableSoftDevice()) {
            updateSoftDeviceEncTmk(cipher.getC1C3C2Bytes());
        } else {
            Device up = new Device();
            up.setId(device.getId());
            up.setEncTmk(cipher.getC1C3C2Hex());
            up.setTmkStatus(TmkStatusEnum.finished.name());
            spDeviceMapper.updateById(up);
        }
        updateTmkInit(true);
    }

    private Device getOneByStatus(TmkStatusEnum status) {
        if (enableSoftDevice()) {
            Device device = new Device();
            device.setManufacturerModel(BouncyCastleProvider.PROVIDER_NAME);
            device.setEncKeyIdx(1);
            device.setServiceIp("127.0.0.1");
            device.setServicePort(8889);
            device.setAccessCredentials("543");
            device.setEncTmk(CodecUtils.encodeHex(getSoftDeviceEncTmk()));
            return device;
        }

        return spDeviceMapper.selectOneByStatus(status);
    }

    private boolean isTmkInit() {
        ParamConf conf = paramConfMapper.selectByKey(ParamConfKeyConstant.TMK_INIT);
        return conf != null && String.valueOf(true).equals(conf.getValue());
    }

    private void updateTmkInit(boolean value) {
        ParamConf conf = paramConfMapper.selectByKey(ParamConfKeyConstant.TMK_INIT);
        if (conf == null) {
            conf = new ParamConf();
            conf.setValue(String.valueOf(value));
            conf.setKey(ParamConfKeyConstant.TMK_INIT);
            conf.setCreatTime(LocalDateTime.now());
            paramConfMapper.insert(conf);
        } else {
            conf.setValue(String.valueOf(value));
            paramConfMapper.updateById(conf);
        }
    }

    private boolean enableSoftDevice() {
        ParamConf conf = paramConfMapper.selectByKey(ParamConfKeyConstant.ENABLE_SOFT_DEVICE);
        return conf != null && String.valueOf(true).equals(conf.getValue());
    }

    private byte[] getSoftDeviceEncTmk() {
        boolean tmkInit = isTmkInit();
        Assert.isTrue(tmkInit, "主密钥未初始化");
        boolean enabled = enableSoftDevice();
        Assert.isTrue(enabled, "未启用软设备");
        ParamConf conf = paramConfMapper.selectByKey(ParamConfKeyConstant.SOFT_ENC_TMK);
        Assert.notNull(conf, "数据异常, 未找到软设备主密钥记录");
        return CodecUtils.decodeBase64(conf.getValue());
    }

    private void updateSoftDeviceEncTmk(byte[] encTmk) {
        ParamConf conf = paramConfMapper.selectByKey(ParamConfKeyConstant.SOFT_ENC_TMK);
        if (conf == null) {
            conf = new ParamConf();
            conf.setValue(CodecUtils.encodeBase64(encTmk));
            conf.setKey(ParamConfKeyConstant.SOFT_ENC_TMK);
            conf.setCreatTime(LocalDateTime.now());
            paramConfMapper.insert(conf);
        } else {
            conf.setValue(CodecUtils.encodeBase64(encTmk));
            paramConfMapper.updateById(conf);
        }
    }

    private void checkName(String name) {
        LambdaQueryWrapper<Device> wrapper = new LambdaQueryWrapper<Device>()
                .eq(Device::getName, name);
        Device exist = spDeviceMapper.selectOne(wrapper);
        Assert.isNull(exist, "设备名称已存在");
    }

}
